#!/usr/bin/env python

# PEP 517 builds do not have . in sys.path
import os
import sys
sys.path.insert(0, os.path.dirname(__file__))

from setuptools import Extension

from sage_setup import sage_setup
from sage.env import cython_aliases

if not (len(sys.argv) > 1 and (sys.argv[1] in ["sdist", "egg_info", "dist_info"])):
    aliases = cython_aliases(required_modules=('Singular',), optional_modules=())
    ext_modules = [Extension("PySingular", ["PySingular.cpp"],
                             extra_compile_args=aliases['SINGULAR_CFLAGS'],
                             include_dirs=aliases['SINGULAR_INCDIR'],
                             libraries=aliases['SINGULAR_LIBRARIES'],
                             library_dirs=aliases['SINGULAR_LIBDIR'],
                             extra_link_args=aliases['SINGULAR_LIBEXTRA'])]
else:
    ext_modules = []

sage_setup(['sagemath-singular'],
           recurse_packages=('sage', 'passagemath_singular'),
           required_modules=('Singular', 'factory', 'givaro', 'gsl'),
           spkgs=['singular'],
           package_data={"sage": [
               "ext_data/singular/**",
           ]},
           ext_modules=ext_modules)



# sage_setup: distribution = sagemath-singular
# distutils: include_dirs = SINGULAR_INCDIR
# distutils: extra_compile_args = SINGULAR_CFLAGS
# distutils: libraries = SINGULAR_LIBRARIES
# distutils: library_dirs = SINGULAR_LIBDIR
# distutils: language = c++
# distutils: extra_compile_args = -std=c++17
"""
Declarations of Singular's C/C++ Functions

.. NOTE::

    Our ``ring``, ``poly``,... types are not the Singular ``ring``,
    ``poly``,... types. They are dereferences. So a Singular ``ring`` is
    a pointer to a ``ring`` from Sage.

AUTHOR:

- Martin Albrecht (2009-07): initial implementation
"""
#*****************************************************************************
#       Copyright (C) 2009 Martin Albrecht <malb@informatik.uni-bremen.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  http://www.gnu.org/licenses/
#*****************************************************************************
from sage.libs.gmp.types cimport mpz_t, mpz_ptr

cdef extern from "factory/factory.h":

    #
    # CF OPTIONS
    #

    void On( int )
    void Off( int )
    int isOn( int )

    cdef int SW_USE_CHINREM_GCD
    cdef int SW_USE_EZGCD
    cdef int SW_USE_NTL
    cdef int SW_USE_NTL_GCD_0
    cdef int SW_USE_NTL_GCD_P
    cdef int SW_USE_NTL_SORT

cdef extern from "singular/Singular/libsingular.h":
    """
    // compatibility for singular 4.3.2p10 and before
    #if SINGULAR_VERSION <= 4330
    #define ringorder_ip ringorder_rp
    #define BIGINTVEC_CMD INTVEC_CMD
    #endif
    """

    #
    # OPTIONS
    #

    cdef unsigned int singular_options "si_opt_1"           # previously 'test'
    cdef unsigned int singular_verbose_options "si_opt_2"   # previously 'verbose'

    # actual options
    cdef int OPT_PROT
    cdef int OPT_REDSB
    cdef int OPT_NOT_BUCKETS
    cdef int OPT_NOT_SUGAR
    cdef int OPT_INTERRUPT
    cdef int OPT_SUGARCRIT
    cdef int OPT_DEBUG
    cdef int OPT_REDTHROUGH
    cdef int OPT_RETURN_SB
    cdef int OPT_FASTHC
    cdef int OPT_OLDSTD
    cdef int OPT_KEEPVARS
    cdef int OPT_STAIRCASEBOUND
    cdef int OPT_MULTBOUND
    cdef int OPT_DEGBOUND
    cdef int OPT_REDTAIL
    cdef int OPT_INTSTRATEGY
    cdef int OPT_INFREDTAIL
    cdef int OPT_SB_1
    cdef int OPT_NOTREGULARITY
    cdef int OPT_WEIGHTM

    # getter/setter functions
    int Sy_bit(int)
    int Sy_inset(int x,int s)
    int BTEST1(int)
    int BVERBOSE(int)

    # ideal flags
    cdef int  FLAG_STD
    cdef int  FLAG_TWOSTD

    #
    # STRUCTS
    #

    # rational numbers
    ctypedef struct number "snumber":
        mpz_t z
        mpz_t n
        int s

    # See singular/libpolys/coeffs/coeffs.h for documentation
    cdef enum n_coeffType:
        n_unknown
        n_Zp
        n_Q
        n_R
        n_GF
        n_long_R
        n_algExt
        n_transExt
        n_long_C
        n_Z
        n_Zn
        n_Znm
        n_Z2m
        n_CF

    ctypedef struct ring "ip_sring"
    ctypedef struct AlgExtInfo

    ctypedef struct n_Procs_s:

        number* cfDiv(number *, number *, const n_Procs_s* r)
        number* cfAdd(number *, number *, const n_Procs_s* r)  # algebraic number addition
        number* cfSub(number *, number *, const n_Procs_s* r)
        number* cfMult(number *, number *, const n_Procs_s* r)  # algebraic number multiplication

        number* (*cfInit)(int i, const n_Procs_s* r ) # algebraic number from int
        number* (*cfInitMPZ)(mpz_t i, const n_Procs_s* r)
        number* (*cfParameter)(int i, const n_Procs_s* r)
        int (*cfParDeg)(number* n, const n_Procs_s* r)
        int (*cfSize)(number* n, const n_Procs_s* r)
        int (*cfInt)(number* n, const n_Procs_s* r)
        int (*cdDivComp)(number* a,number* b, const n_Procs_s* r)
        number* (*cfGetUnit)(number* a, const n_Procs_s* r)
        number* (*cfExtGcd)(number* a, number* b, number* *s, number* *t , const n_Procs_s* r)

        void (*cfDelete)(number **, const n_Procs_s*)

        number* (*cfInpNeg)(number* a,  const n_Procs_s* r)
        number* (*cfInvers)(number* a,  const n_Procs_s* r)
        number* (*cfCopy)(number* a,  const n_Procs_s* r) # deep copy of algebraic number
        number* (*cfRePart)(number* a, const n_Procs_s* cf)
        number* (*cfImPart)(number* a, const n_Procs_s* cf)
        void (*cfWrite)(number* a, const n_Procs_s* r)
        void (*cfNormalize)(number* a,  const n_Procs_s* r)

        bint (*cfDivBy)(number* a, number* b, const n_Procs_s* r)
        bint (*cfGreater)(number* a, number* b, const n_Procs_s* )
        bint (*cfEqual)(number* a,number* b, const n_Procs_s* )
        bint (*cfIsZero)(number* a, const n_Procs_s* ) # algebraic number comparison with zero
        bint (*cfIsOne)(number* a, const n_Procs_s* )  # algebraic number comparison with one
        bint (*cfIsMOne)(number* a, const n_Procs_s* )
        bint (*cfGreaterZero)(number* a, const n_Procs_s* )
        void (*cfPower)(number* a, int i, number* * result,  const n_Procs_s* r) # algebraic number power

        ring *extRing
        int ch
        mpz_ptr modBase
        unsigned long modExponent

        # n_coeffType type
        int type

    # polynomials

    const char ** n_ParameterNames(const n_Procs_s* r)

    int n_NumberOfParameters(const n_Procs_s* r)

    ctypedef struct poly "spolyrec":
        poly *next
        number *coef
        unsigned long exp[1]

    # ideals

    ctypedef struct ideal "sip_sideal":
        poly **m # gens array
        long rank # rank of module, 1 for ideals
        int nrows # always 1
        int ncols # number of gens

    # polynomial procs
    ctypedef struct p_Procs_s "p_Procs_s":
        pass
    # rings

    ctypedef struct ring "ip_sring":
        int  *order  # array of orderings
        int  *block0 # starting pos
        int  *block1 # ending pos
        int  **wvhdl # weight vectors
        int  OrdSgn  # 1 for polynomial rings
        int  ShortOut # control printing
        int  CanShortOut # control printing capabilities
        number *minpoly # minpoly for base extension field
        char **names # variable names
        p_Procs_s *p_Procs #polxnomial procs
        ideal *qideal #quotient ideal

        short N # number of variables

        int pCompIndex # index of components
        unsigned long bitmask # mask for getting single exponents

        n_Procs_s* cf  # coefficient field/ring
        int ref

        # return total degree of p

        long (*pLDeg)(poly *p, int *l, ring *r)
        long (*pLDegOrig)(poly *p, int *l, ring *r)
        long (*pFDeg)(poly *p, ring *r)
        long (*pFDegOrig)(poly *p, ring *r)

    # a bunch of functions for degree

    long p_FDeg(poly *p, ring *r)
    long p_LDeg(poly *p, int *l, ring *r)
    long p_Deg(poly *p, ring *r)
    long p_WTotaldegree(poly *p, ring *r)
    long p_Totaldegree(poly *p, ring *r)
    long p_WDegree(poly *p, ring *r)

    # available ring orders

    ctypedef struct AlgExtInfo:
        ring * r

    cdef enum rRingOrder_t:
        ringorder_no
        ringorder_a
        ringorder_a64 # for int64 weights
        ringorder_c
        ringorder_C
        ringorder_M
        ringorder_S
        ringorder_s
        ringorder_lp
        ringorder_dp
        ringorder_ip
        ringorder_Dp
        ringorder_wp
        ringorder_Wp
        ringorder_ls
        ringorder_ds
        ringorder_Ds
        ringorder_ws
        ringorder_Ws
        ringorder_L

    # groebner basis options

    cdef enum tHomog:
        isNotHomog
        isHomog
        testHomog

    # ideals

    ctypedef struct ideal "sip_sideal":
        poly **m # gens array
        long rank # rank of module, 1 for ideals
        int nrows # always 1
        int ncols # number of gens

    # dense matrices

    ctypedef struct matrix "ip_smatrix":
        poly **m # gens array
        long rank # rank of module, 1 for normal matrices
        int nrows # number of rows
        int ncols # number of columns

    # integer vectors

    cdef cppclass intvec:
        intvec()
        intvec(int, int, int)
        int *(*ivGetVec)() # this is internal actually
        int (*rows)()
        int (*cols)()
        int (*length)()
        int (*resize)(size_t s)
        int (*get "operator[]")(int i)
        int row
        int col

    cdef cppclass bigintmat:
        int (*length)()
        number* (*get)(int i)

    # omalloc bins

    ctypedef struct omBin "omBin_s"

    # pairs and polys for Groebner Strategy objects

    ctypedef struct LObject:
        pass

    ctypedef struct TObject:
        pass

    # Groebner Strategy objects

    ctypedef struct skStrategy:
        int ak # free module rank
        int sl # last index (inclusive)
        int (*red)(LObject * L,skStrategy *strat)
        void (*initEcart)(LObject *L)
        int (*posInT)(TObject *T,int tl,LObject h) # return the required position in T
        int (*posInL)(LObject *set, int length, LObject* L, skStrategy *strat) # return the required position in L
        void (*enterS)(LObject h, int pos, skStrategy *strat, int atR) # enter into S
        void (*initEcartPair)(LObject * h, poly *f, poly *g, int ecartF, int ecartG)
        int (*posInLOld)(LObject * Ls, int Ll, LObject* Lo, skStrategy *strat)
        LObject P
        ideal *Shdl
        ideal *D
        ideal *M
        poly ** S
        int * ecartS
        int * lenS
        int * fromQ
        unsigned long* sevS
        unsigned long* sevT
        TObject *T
        LObject *L
        LObject *B
        poly* kHEdge
        poly* kNoether
        poly* t_kHEdge
        poly* kNoetherTail()
        poly* t_kNoether
        bint *NotUsedAxis
        bint *pairtest
        void *R
        int *S_2_R
        bint noTailReduction

    ctypedef struct data_union:
        ring *uring

    # interpreter objects

    ctypedef struct attr "sattr":
        void (*Init)()
        char * name
        void * data
        attr * next
        int atyp # the type of the attribute, describes the data field

        void (*Print)()
        attr *(*Copy)() # copy all arguments
        void *(*CopyA)()# copy the data of this attribute
        attr *(*set)(char * s, void * data, int t)
        attr *(*get)(char * s)
        void (*kill)()
        void (*killAll)()

    ctypedef struct idhdl "idrec":
        idhdl *next
        char *id
        int typ
        short lev
        short ref
        data_union data

    ctypedef struct leftv "sleftv":
        leftv *next
        char *id
        void* data
        # data is some union, so this might be very dangerous, but I am lazy now
        attr *attribute
        void (* Copy)(leftv*)
        void (* Init)()
        void (* CleanUp)(ring *r)
        int  rtyp

    ctypedef struct package "ip_package":
        int language
        idhdl *idroot

    #
    # GLOBAL VARIABLES
    #

    # current ring

    cdef ring *currRing
    cdef ideal *currQuotient
    # omalloc bin for numbers

    cdef omBin *rnumber_bin

    # omalloc bin for rings

    cdef omBin *sip_sring_bin

    # omalloc bin for lists

    cdef omBin *slists_bin

    # integer conversion constant

    cdef long SR_INT

    cdef package *basePack
    cdef package *currPack

    cdef idhdl *basePackHdl
    cdef idhdl *currPackHdl
    cdef idhdl *currRingHdl

    cdef int errorreported
    cdef int si_opt_2    #  previously 'verbose'
    cdef void * currentVoice
    cdef int myynest
    cdef int printlevel

    ctypedef char * const_char_ptr "const char *"
    cdef extern void (*WerrorS_callback)(const_char_ptr)

    #
    # FUNCTIONS
    #

    # singular init

    int siInit(char *)

    ctypedef short (*cfInitCharProc)(coeffs, void *)

    n_coeffType nRegister(n_coeffType n, cfInitCharProc p)

    # external resource init

    void feInitResources(char *name)

    # external resource query

    char* feGetResource(const char id)

    void *omAlloc(size_t size)

    # calloc

    void *omAlloc0(size_t size)

    # typed malloc from bin

    void *omAllocBin(omBin *bin)
    void omFreeBin(void* foo, void* bin)

    # typed calloc from bin

    void *omAlloc0Bin(omBin *bin)

    # strdup

    char *omStrDup(char *)

    # free

    void omFree(void *)

    void omfree(void *)

    # construct ring with characteristic, number of vars and names

    ring *rDefault(int char , int nvars, char **names)
    ring *rDefault(const n_Procs_s* cf, int nvars, char **names)
    ring *rDefault(int ch             , int nvars, char **names,int ord_size, rRingOrder_t *ord, int *block0, int *block1, int **wvhdl)
    ring *rDefault(const n_Procs_s* cf, int nvars, char **names,int ord_size, rRingOrder_t *ord, int *block0, int *block1, int **wvhdl)

    # see coeffs.h
    ctypedef struct  GFInfo:
        int GFChar
        int GFDegree
        const char* GFPar_name

    # parameter is pointer to gGFInfo
    #
    n_Procs_s* nInitChar(n_coeffType t, void * parameter)

    # ring destructor

    void rDelete(ring *r)

    # return characteristic of r

    int rChar(ring *r)

    # return name of the i-th variable of ring r

    char* rRingVar(short i, ring *r)

    # before changing a ring struct, call this

    void rUnComplete(ring *r)

    # after changing a ring struct, call this, computes internal
    # representation for monomials etc.
    void rComplete(ring *r, int force)

    # deep copy of ring

    ring *rCopy0(ring *)

    # change current ring to r

    void rChangeCurrRing(ring *r)

    # return True if ring has components

    int rRing_has_Comp(ring *r)

    int rHasGlobalOrdering(ring *r)
    int rHasLocalOrMixedOrdering(ring *r)

    # return new empty monomial

    poly *p_Init(ring *r)

    # Returns new monomial with exp vector copied from Lm(p)
    poly *p_LmInit(poly *p, ring *r)

    # return constant polynomial from int

    poly *p_ISet(int i, ring *r)

    # return constant polynomial from number

    poly *p_NSet(number *n,ring *r)

    # destructor for polynomials

    void p_Delete(poly **p, ring *r)

    # set the coefficient n for the current list element p in r

    int p_SetCoeff(poly *p, number *n, ring *r)

    # set the coefficient n for the current list element p in r
    # without deleting the current coefficient first

    int p_SetCoeff0(poly *p, number *n, ring *r)

    # get the coefficient of the current list element p in r

    number *p_GetCoeff(poly *p, ring *r)

    # get the module component
    int p_GetComp(poly *p, ring *r)

    # set the module component
    void p_SetComp(poly *p, int v, ring *r)

    # set the exponent e at index v for the monomial p, v starts at 1

    int p_SetExp(poly *p, int v, int e, ring *r)

    # get the exponent at index v of the monomial p in r, v starts at 1

    int p_GetExp(poly *p, int v, ring *r)

    # get the maximal exponent in p

    unsigned long p_GetMaxExp(poly *p, ring *r)

    # if SetExp is called on p, p_Setm needs to be called afterwards to finalize the change.

    void p_Setm(poly *p, ring *r)

    # gets a component out of a polynomial vector

    poly *pTakeOutComp(poly **, int)

    # deep copy p

    poly *p_Copy(poly *p, ring *r)

    # homogenizes p by multiplying certain powers of the varnum-th variable

    poly *p_Homogen (poly *p, int varnum, ring *r)

    # return whether a polynomial is homogeneous

    int p_IsHomogeneous(poly *p, const  ring *r)

    # return string representation of p

    char *p_String(poly *p, ring *r, ring *r)

    # normalize p, needed e.g. for polynomials over the rationals

    void p_Normalize(poly *p, ring *r)

    # makes it so that leading coefficient == 1
    void pNorm(poly *p)

    # return -p, p is destroyed

    poly *p_Neg(poly *p, ring *r)

    # return p*n, p is const (i.e. copied)

    poly *pp_Mult_nn(poly *p, number *n, ring *r)

    # return p*m, does neither destroy p nor m

    poly *pp_Mult_mm(poly *p, poly *m, ring *r)

    # return p+q, destroys p and q

    poly *p_Add_q(poly *p, poly *q, ring *r)

    # return p - m*q, destroys p; const: q,m

    poly *p_Minus_mm_Mult_qq(poly *p, poly *m, poly *q, ring *r)

    # return p + m*q destroys p, const: q, m

    poly *p_Plus_mm_Mult_qq(poly *p, poly *m, poly *q, ring *r)

    # return p*q, does neither destroy p nor q
    poly *pp_Mult_qq(poly *p, poly *q, ring *r)

    # return p*q, destroys p and q
    poly *p_Mult_q(poly *p, poly *q, ring *r)

    # polynomial division, ignoring the remainder
    # via singclap_pdivide resp. idLift, destroys p,q
    poly *p_Divide(poly *p, poly *q, ring *r)

    # divide monomial p by monomial q, p,q const

    poly *pMDivide(poly *p,poly *q)
    poly *p_MDivide(poly *p, poly *q, ring *r)

    # return the i-th power of p; p destroyed, requires global ring

    poly *pPower(poly *p, int exp)

    # return new copy of lm(p), coefficient copied, next=NULL, p may be NULL

    poly *p_Head(poly *p, ring *r)

    # return TRUE, if leading monom of a divides leading monom of b
    # i.e., if there exists a expvector c > 0, s.t. b = a + c

    int p_DivisibleBy(poly *a, poly *b, ring *r)

    # like pDivisibleBy, except that it is assumed that a!=NULL, b!=NULL

    int p_LmDivisibleBy(poly *a, poly *b, ring *r)

    # least common multiplies for monomials only, result is written to m
    # p_Setm must be called on m afterwards.

    void pLcm(poly *a, poly *b, poly *m)

    # total degree of p

    long p_Totaldegree(poly *p, ring *r)

    long pLDeg1_Totaldegree(poly * p,int *l, ring * r)

    # iterate through the monomials of p

    poly *pNext(poly *p)

    # Returns the number of monomials in the poly
    int pLength(poly *a)

    # compare l and r

    int p_Cmp(poly *l, poly *r, ring *r)

    # compare exponent vectors only

    int p_ExpVectorEqual(poly *p, poly *m, ring *r)

    # TRUE if poly is constant

    int p_IsConstant(poly *, ring *)

    # like p_IsConstant but p must be !=NULL

    int p_LmIsConstant(poly *p, ring *)

    # TRUE if poly is unit

    int p_IsUnit(poly *p, ring *r)

    # TRUE if poly  is one

    bint p_IsOne(const poly *p, const ring *r)

    # substitute monomial for variable given by varidx in poly

    poly *pSubst(poly *p, int varidx, poly *value)

    # inverse of poly, if possible

    poly *p_Series(int n, poly *, poly *, intvec *, ring *r)

    # gcd of f and g

    poly *singclap_gcd ( poly *f, poly *g, ring * r )

    # resultant of f and g in x

    poly *singclap_resultant ( poly *f, poly *g , poly *x, ring * r)

    # extended gcd of f and g

    int singclap_extgcd( poly *f, poly *g, poly *res, poly *pa, poly *pb, ring * r )

    # full polynomial division (as opposed to monomial division)

    poly *singclap_pdivide ( poly *f, poly *g, ring * r )

    # factorization

    ideal *singclap_factorize ( poly *f, intvec ** v , int with_exps, ring * r)

    # return determinant of i
    poly *singclap_det(matrix *i, ring * r)

    # normal form calculation of p with respect to i, q is quotient
    # ring.

    poly *kNF(ideal *i, ideal *q, poly *p)
    poly *kNF(ideal *i, ideal *q, poly *p, int syzComp, int lazyReduce)

    # derive p with respect to i-th variable

    poly *pDiff(poly *p, int i)

    # TRUE if p is a vector

    int pIsVector(poly *p)

    # return current component level

    int pGetComp(poly *p)

    # general number constructor

    number *n_Init(int n, n_Procs_s *cf)

    # general number destructor

    void n_Delete(number **n, n_Procs_s *cf)

    # Copy this number
    number *n_Copy(number *n, n_Procs_s *cf)

    # Invert this number
    int n_IsUnit(number *n, const n_Procs_s *cf)
    number *n_Invers(number *n, const n_Procs_s *cf)

    # Characteristic of coefficient domain
    int n_GetChar(const n_Procs_s *cf)

    # rational number from int

    number *nlInit(int)

    # rational number from int

    number *nlRInit(int)

    # rational number from numerator and denominator

    number *nlInit2(int i, int j,const n_Procs_s* cf)

    # simplify rational number (cancel common factors)

    number *nlNormalize(number *)

    # copy a number

    number *nlCopy(number *)

    # number to integer handle

    long SR_TO_INT(number *)

    # mpz_t to integer handle

    long SR_HDL(number *)

    # get C int from ZmodN
    int nrnInt(number *v)

    # ideal constructor

    ideal *idInit(int size, int rank)

    # ideal destructor

    void id_Delete(ideal **, ring *)

    # lifting

    ideal *idLift(ideal *mod, ideal *submod, ideal **rest, int goodShape, int isSB, int divide)

    # number of generators of ideal

    int IDELEMS(ideal *i)

    # remove zero entries

    void idSkipZeroes (ideal *ide)

    # rank of free module for m

    long id_RankFreeModule(ideal *m, ring *r)

    # buchberger's algorithm

    ideal *kStd(ideal *i, ideal *q, tHomog h, intvec *w)

    # slimgb algorithm

    ideal *t_rep_gb(ring *r,ideal *arg_I, int syz_comp, int F4_mode)

    # interreduction

    ideal *kInterRed(ideal *i, ideal *q)

    # TRUE if ideal is module

    int idIsModule(ideal *m, ring *r)

    # convert module to matrix

    matrix *idModule2Matrix(ideal *i)

    # convert matrix to module

    ideal * idMatrix2Module(matrix *m)

    # dense matrix constructor

    matrix *mpNew(int i, int j)

    # gauss-bareiss algorithm

    void smCallBareiss(ideal *, int, int, ideal *, intvec**)

    # determinant
    poly *smCallDet(ideal *)

    # check which det algorithm to choose
    int smCheckDet(ideal *, int, int)

    # get element at row i and column j

    poly *MATELEM(matrix *, int i, int j)

    # number columns of matrix

    int MATCOLS(matrix *)

    # number for rows of matrix

    int MATROWS(matrix *)

    # Groebner Strategy functions

    void initBuchMoraCrit(skStrategy *strat)
    void initEcartNormal (LObject* h)
    void initEcartBBA (LObject* h)
    void initEcartPairBba (LObject* Lp,poly *f,poly *g,int ecartF,int ecartG)

    # init strat with F and Q
    void initS (ideal *F, ideal *Q,skStrategy *strat)

    void enterL (LObject *set,int *length, int *LSetmax, LObject p,int at)
    void enterSBba (LObject p,int atS,skStrategy *strat, int atR)

    skStrategy * new_skStrategy "new skStrategy"()
    void * delete_skStrategy "delete "(skStrategy *doomed) # needs currRing

    # test validity of strat
    void kTest(skStrategy *strat)

    # head reduction
    poly *redNF(poly *p, int index, int nonorm, skStrategy *strat)
    # tail reduction
    poly *redtailBba(poly *p, int index, skStrategy *strat)

    cdef int CMD_1
    cdef int CMD_2
    cdef int CMD_12
    cdef int CMD_3
    cdef int CMD_13
    cdef int CMD_23
    cdef int CMD_123
    cdef int CMD_M
    cdef int ROOT_DECL
    cdef int ROOT_DECL_LIST
    cdef int RING_DECL
    cdef int RING_DECL_LIST
    cdef int INT_CMD
    cdef int INTMAT_CMD
    cdef int POLY_CMD
    cdef int PROC_CMD
    cdef int RING_CMD
    cdef int QRING_CMD
    cdef int STRING_CMD
    cdef int VECTOR_CMD
    cdef int IDEAL_CMD
    cdef int MODUL_CMD
    cdef int NUMBER_CMD
    cdef int MATRIX_CMD
    cdef int LIST_CMD
    cdef int INTVEC_CMD
    cdef int BIGINTVEC_CMD
    cdef int NONE
    cdef int RESOLUTION_CMD
    cdef int PACKAGE_CMD

    cdef int V_SHOW_MEM
    cdef int V_YACC
    cdef int V_REDEFINE
    cdef int V_READING
    cdef int V_LOAD_LIB
    cdef int V_DEBUG_LIB
    cdef int V_LOAD_PROC
    cdef int V_DEF_RES
    cdef int V_SHOW_USE
    cdef int V_IMAP
    cdef int V_PROMPT
    cdef int V_NSB
    cdef int V_CONTENTSB
    cdef int V_CANCELUNIT
    cdef int V_DEG_STOP

    #
    # INTERPRETER
    #
    leftv iiRETURNEXPR

    cdef omBin* sleftv_bin

    int     IsCmd(char *n, int  tok)

    idhdl* ggetid(char *n)

    bint iiMake_proc(idhdl *pn, package *pack, leftv *sl)

    bint iiExprArith1(leftv *res, leftv* a, int op)
    bint iiExprArith2(leftv *res, leftv* a, int op, leftv *b, bint proc_call)
    bint iiExprArith3(leftv *res, int op, leftv *a, leftv *b, leftv *c)
    bint iiExprArithM(leftv *res, leftv* a, int op)

    bint iiLibCmd(char *, bint autoexport, bint tellerror, bint force)

    package *IDPACKAGE(idhdl *)

    cdef idhdl *IDROOT

    cdef idhdl *IDNEXT(idhdl *)

    cdef int IDTYP(idhdl *)

    idhdl *enterid(char * a, int lev, int t, idhdl** root, bint init)

    void atSet(leftv *root, char * name, void * data, int typ)
    void *atGet(leftv *root, char *name, int typ)

    int hasFlag(leftv *A, int F)
    void setFlag(leftv *A, int F)
    void resetFlag(leftv *A, int F)

    ctypedef number* (*nMapFunc)(number *c, const n_Procs_s* src, const n_Procs_s* dst)

cdef extern from "singular/coeffs/coeffs.h":

    number *ndCopyMap(number *, const n_Procs_s* src,const n_Procs_s* dst)

    ctypedef struct LongComplexInfo:
        short float_len
        short float_len2
        const char* par_name

cdef extern from "singular/coeffs/rmodulo2m.h":

    #init 2^m from a long
    number *nr2mMapZp(number *,const n_Procs_s* src,const n_Procs_s* dst)

cdef extern from "singular/coeffs/shortfl.h":
    """
    static inline SI_FLOAT sage_nrFloat(number n) { // copy from shortfl.cc to allow inlining
        SI_FLOAT f = 0;
        memcpy(&f, &n, sizeof(f) < sizeof(n) ? sizeof(f) : sizeof(n));
        return f;
    }

    static inline number sage_nrInit(SI_FLOAT f) {
        number n = 0;
        memcpy(&n, &f, sizeof(f) < sizeof(n) ? sizeof(f) : sizeof(n));
        return n;
    }
    """
    ctypedef double SI_FLOAT  # actually might be double or float
    SI_FLOAT nrFloat "sage_nrFloat" (number *n)
    number *sage_nrInit(SI_FLOAT)

cdef extern from "singular/kernel/maps/gen_maps.h":

    # mapping from p in r1 by i2 to r2

    poly *maMapPoly(poly *p, ring *r1, ideal *i2, ring *r2, const nMapFunc nMap)

cdef extern from "singular/kernel/maps/fast_maps.h":

    # mapping from ideal i1 in r1 by i2 to r2

    ideal *fast_map_common_subexp(ideal *i1, ring *r1, ideal *i2, ring *r2)

cdef extern from "singular/polys/ext_fields/algext.h":

    naInitChar(n_Procs_s* cf, void * infoStruct)

    nMapFunc naSetMap(const n_Procs_s* src, const n_Procs_s* dst)

cdef extern from "singular/coeffs/mpr_complex.h":
    cdef cppclass gmp_float:
        gmp_float(double v)
        double to_double "operator double"() const

    cdef cppclass gmp_complex:
        gmp_complex(double re, double im)
        gmp_float real() const
        gmp_float imag() const

    char *floatToStr(const gmp_float & r, const unsigned int oprec)
    char *complexToStr(gmp_complex & c, const unsigned int oprec, const n_Procs_s* src)


cdef extern from "singular/coeffs/rmodulon.h":

    # init ZmodN from GMP
    number *nrnMapGMP(number *v,const n_Procs_s* src,const n_Procs_s* dst)

    nMapFunc nrnSetMap(const n_Procs_s* src,const n_Procs_s* dst)

cdef extern from "singular/coeffs/rmodulon.h":
    # see rmodulon.h

    ctypedef struct ZnmInfo:
       mpz_ptr base
       unsigned long exp

cdef extern from "singular/coeffs/rintegers.h":

    # init integer
    number *nrzInit(int i, const n_Procs_s* cf)

cdef extern from "singular/polys/weight.h":

    double wFunctionalBuch(int *degw, int *lpol, int npol, double *rel, double wx, double wNsqr)

cdef extern from "singular/polys/prCopy.h":
    poly *prCopyR_NoSort(poly *p, ring *r, ring *dest_r)
    poly *prCopyR(poly *p, ring *r, ring *dest_r)

    cdef int LANG_TOP

cdef extern from "singular/polys/sbuckets.h":
    #sBucket is actually a class, but we handle it opaquely, so we call it a "struct" here.
    ctypedef struct sBucket:
        pass

    #create an sBucket
    sBucket *sBucketCreate(ring *r)

    #destroy an sBucket (note: pointer to pointer)
    void sBucketDestroy(sBucket **bucket)

    #merge contents of sBucket into polynomial and clear bucket
    #(use when monomials are distinct).
    #assumes length <= 0 || pLength(p) == length
    void sBucketClearMerge(sBucket *bucket, poly **p, int *length)

    #add contents of sBucket into polynomial and clear bucket
    #(can handle repeated monomials)
    void sBucketClearAdd(sBucket *bucket, poly **p, int *length)

    #inline versions that in addition clear the pointer bucket afterwards
    void sBucketDestroyMerge(sBucket *bucket, poly **p, int *length)
    void sBucketDestroyAdd(sBucket *bucket, poly *p, int *length)

    #delete bucket constant and clear pointer
    void sBucketDeleteAndDestroy(sBucket **bucket_pt)

    #merge p into bucket (distinct monomials assumed)
    #destroys poly in the process
    void sBucket_Merge_p(sBucket *bucket, poly *p, int lp)

    #merge p into bucket  (distinct monomials assumed)
    #destroys poly in the process
    void sBucket_Merge_m(sBucket *bucket, poly *p)

    #adds p into bucket (distinct monomials assumed)
    #destroys poly in the process
    void sBucket_Add_p(sBucket *bucket, poly *p, int lp)

    #adds p into bucket (distinct monomials assumed)
    #destroys poly in the process
    void sBucket_Add_m(sBucket *bucket, poly *p)

    #sorts p with bucketSort: assumes all monomials of p are different
    poly *sBucketSortMerge(poly *p, const ring *r)

    #sorts p with bucketSort: p may have equal monomials
    poly *sBucketSortAdd(poly *p, const ring *r)

cdef extern from "singular/polys/nc/nc.h":
    # Non-commutative functions
    ctypedef enum nc_type:
        nc_error
        nc_general
        nc_skew
        nc_comm
        nc_lie
        nc_undef
        nc_exterior

    void ncRingType(ring *, nc_type)
    nc_type ncRingType_get "ncRingType" (ring *)
    int nc_CallPlural(matrix* CC, matrix* DD, poly* CN, poly* DN, ring* r)
    bint nc_SetupQuotient(ring *, ring *, bint)

cdef extern from "singular/coeffs/longrat.h":

    # get numerator

    number *nlGetNumerator(number *n, const n_Procs_s* cf)

    # get denominator

    number *nlGetDenom(number *n, const n_Procs_s* cf)

    # rational number from numerator and denominator

    number *nlInit2gmp(mpz_t n, mpz_t d,const n_Procs_s* cf)

    # delete rational number

    void nlDelete(number **n, const n_Procs_s* cf)

cdef extern from "singular/polys/nc/sca.h":
    void sca_p_ProcsSet(ring *, p_Procs_s *)
    void scaFirstAltVar(ring *, int)
    void scaLastAltVar(ring *, int)

cdef extern from "singular/polys/monomials/ring.h":
    bint rIsPluralRing(ring* r)
    void rPrint "rWrite"(ring* r)
    char* rOrderingString "rOrdStr"(ring* r)
    void pDebugPrint "p_DebugPrint" (poly*p, ring* r)

cdef extern from "singular/kernel/combinatorics/stairc.h":
    # Computes the monomial basis for R[x]/I
    ideal *scKBase(int deg, ideal *s, ideal *Q)

cdef extern from "singular/Singular/lists.h":
    ctypedef struct lists "slists":
        int    nr
        leftv  *m
        void (*Init)(int n)

cdef extern from "singular/kernel/GBEngine/kstd1.h":
    cdef extern int Kstd1_deg   # degBound, default 0
    cdef extern int Kstd1_mu    # multBound, default 0

cdef extern from "singular/kernel/ideals.h":
    ctypedef ideal * resolvente

cdef extern from "singular/kernel/GBEngine/syz.h":
    ctypedef struct syStrategy "ssyStrategy":
        resolvente fullres;
        resolvente minres;
        int length;
        int regularity;
        short list_length;
        short references

cdef extern from "singular/polys/ext_fields/transext.h":
    ctypedef struct TransExtInfo:
        ring * r



# sage_setup: distribution = sagemath-singular
"""
Singular's Groebner Strategy Objects

AUTHORS:

- Martin Albrecht (2009-07): initial implementation
- Michael Brickenstein (2009-07): initial implementation
- Hans Schoenemann (2009-07): initial implementation
"""

#*****************************************************************************
#       Copyright (C) 2009 Martin Albrecht <M.R.Albrecht@rhul.ac.uk>
#       Copyright (C) 2009 Michael Brickenstein <brickenstein@mfo.de>
#       Copyright (C) 2009 Hans Schoenemann <hannes@mathematik.uni-kl.de>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
#*****************************************************************************

cdef extern from *: # hack to get at cython macro
    int unlikely(int)
    int likely(int)

from sage.structure.richcmp cimport richcmp
from sage.libs.singular.decl cimport ideal, ring, poly, currRing
from sage.libs.singular.decl cimport rChangeCurrRing
from sage.libs.singular.decl cimport new_skStrategy, delete_skStrategy, id_RankFreeModule
from sage.libs.singular.decl cimport initEcartBBA, enterSBba, initBuchMoraCrit, initS, pNorm, id_Delete, kTest
from sage.libs.singular.decl cimport omfree, redNF, p_Copy, redtailBba

from sage.libs.singular.ring cimport singular_ring_reference, singular_ring_delete

from sage.rings.polynomial.multi_polynomial_ideal import MPolynomialIdeal, NCPolynomialIdeal
from sage.rings.polynomial.multi_polynomial_ideal_libsingular cimport sage_ideal_to_singular_ideal
from sage.rings.polynomial.multi_polynomial_libsingular cimport MPolynomial_libsingular, MPolynomialRing_libsingular, new_MP
from sage.rings.polynomial.plural cimport new_NCP

cdef class GroebnerStrategy(SageObject):
    """
    A Wrapper for Singular's Groebner Strategy Object.

    This object provides functions for normal form computations and
    other functions for Groebner basis computation.

    ALGORITHM:

    Uses Singular via libSINGULAR
    """
    def __cinit__(self, L):
        """
        Create a new :class:`GroebnerStrategy` object for the
        generators of the ideal ``L``.

        INPUT:

        - ``L`` -- a multivariate polynomial ideal

        EXAMPLES::

            sage: from sage.libs.singular.groebner_strategy import GroebnerStrategy
            sage: P.<x,y,z> = PolynomialRing(QQ)
            sage: I = Ideal([x+z,y+z+1])
            sage: strat = GroebnerStrategy(I); strat
            Groebner Strategy for ideal generated by 2 elements
            over Multivariate Polynomial Ring in x, y, z over Rational Field

        TESTS::

            sage: from sage.libs.singular.groebner_strategy import GroebnerStrategy
            sage: strat = GroebnerStrategy(None)
            Traceback (most recent call last):
            ...
            TypeError: First parameter must be a multivariate polynomial ideal.

            sage: P.<x,y,z> = PolynomialRing(QQ,order='neglex')
            sage: I = Ideal([x+z,y+z+1])
            sage: strat = GroebnerStrategy(I)
            Traceback (most recent call last):
            ...
            NotImplementedError: The local case is not implemented yet.

            sage: P.<x,y,z> = PolynomialRing(CC,order='neglex')
            sage: I = Ideal([x+z,y+z+1])
            sage: strat = GroebnerStrategy(I)
            Traceback (most recent call last):
            ...
            TypeError: First parameter's ring must be multivariate polynomial ring via libsingular.

            sage: P.<x,y,z> = PolynomialRing(ZZ)
            sage: I = Ideal([x+z,y+z+1])
            sage: strat = GroebnerStrategy(I)
            Traceback (most recent call last):
            ...
            NotImplementedError: Only coefficient fields are implemented so far.

        Check that :issue:`27508` is fixed::

            sage: R2.<x,y> = PolynomialRing(QQ, 2, order='lex')
            sage: I2 = R2.ideal(["x^2 - x", "y^2 - y"])
            sage: R2("x^2 + y").mod(I2), R2("x + y^2").mod(I2)
            (x + y, x + y)
        """
        if not isinstance(L, MPolynomialIdeal):
            raise TypeError("First parameter must be a multivariate polynomial ideal.")

        if not isinstance(L.ring(), MPolynomialRing_libsingular):
            raise TypeError("First parameter's ring must be multivariate polynomial ring via libsingular.")

        self._ideal = L

        cdef MPolynomialRing_libsingular R = <MPolynomialRing_libsingular>L.ring()
        self._parent = R
        self._parent_ring = singular_ring_reference(R._ring)

        if not R.term_order().is_global():
            raise NotImplementedError("The local case is not implemented yet.")

        if not R.base_ring().is_field():
            raise NotImplementedError("Only coefficient fields are implemented so far.")

        if (R._ring != currRing):
            rChangeCurrRing(R._ring)

        cdef ideal *i = sage_ideal_to_singular_ideal(L)
        self._strat = new_skStrategy()

        self._strat.ak = id_RankFreeModule(i, R._ring)
        #- creating temp data structures
        initBuchMoraCrit(self._strat)
        self._strat.initEcart = initEcartBBA
        self._strat.enterS = enterSBba
        #- set S
        self._strat.sl = -1
        #- init local data struct
        initS(i, NULL, self._strat)
        self._strat.noTailReduction = False

        cdef int j
        cdef bint base_ring_is_field = R.base_ring().is_field()
        if (R._ring != currRing):
            rChangeCurrRing(R._ring)
        if base_ring_is_field:
            for j in range(self._strat.sl, -1, -1):
                pNorm(self._strat.S[j])

        id_Delete(&i, R._ring)
        kTest(self._strat)

    def __dealloc__(self):
        """
        TESTS::

            sage: from sage.libs.singular.groebner_strategy import GroebnerStrategy
            sage: P.<x,y,z> = PolynomialRing(GF(32003))
            sage: I = Ideal([x + z, y + z])
            sage: strat = GroebnerStrategy(I)
            sage: del strat
        """
        # WARNING: the Cython class self._parent is no longer accessible!
        # see https://github.com/sagemath/sage/issues/11339
        cdef ring *oldRing = NULL
        if self._strat:
            omfree(self._strat.sevS)
            omfree(self._strat.ecartS)
            omfree(self._strat.T)
            omfree(self._strat.sevT)
            omfree(self._strat.R)
            omfree(self._strat.S_2_R)
            omfree(self._strat.L)
            omfree(self._strat.B)
            omfree(self._strat.fromQ)
            id_Delete(&self._strat.Shdl, self._parent_ring)

            if self._parent_ring != currRing:
                oldRing = currRing
                rChangeCurrRing(self._parent_ring)
                delete_skStrategy(self._strat)
                rChangeCurrRing(oldRing)
            else:
                delete_skStrategy(self._strat)
        if self._parent_ring:
            singular_ring_delete(self._parent_ring)

    def _repr_(self) -> str:
        """
        TESTS::

            sage: from sage.libs.singular.groebner_strategy import GroebnerStrategy
            sage: P.<x,y,z> = PolynomialRing(GF(32003))
            sage: I = Ideal([x + z, y + z])
            sage: strat = GroebnerStrategy(I)
            sage: strat # indirect doctest
            Groebner Strategy for ideal generated by 2 elements over
            Multivariate Polynomial Ring in x, y, z over Finite Field of size 32003
        """
        return "Groebner Strategy for ideal generated by %d elements over %s" % (self._ideal.ngens(), self._parent)

    def ideal(self):
        """
        Return the ideal this strategy object is defined for.

        EXAMPLES::

            sage: from sage.libs.singular.groebner_strategy import GroebnerStrategy
            sage: P.<x,y,z> = PolynomialRing(GF(32003))
            sage: I = Ideal([x + z, y + z])
            sage: strat = GroebnerStrategy(I)
            sage: strat.ideal()
            Ideal (x + z, y + z) of Multivariate Polynomial Ring in x, y, z over Finite Field of size 32003
        """
        return self._ideal

    def ring(self):
        """
        Return the ring this strategy object is defined over.

        EXAMPLES::

            sage: from sage.libs.singular.groebner_strategy import GroebnerStrategy
            sage: P.<x,y,z> = PolynomialRing(GF(32003))
            sage: I = Ideal([x + z, y + z])
            sage: strat = GroebnerStrategy(I)
            sage: strat.ring()
            Multivariate Polynomial Ring in x, y, z over Finite Field of size 32003
        """
        return self._parent

    def __richcmp__(self, other, op):
        """
        EXAMPLES::

            sage: from sage.libs.singular.groebner_strategy import GroebnerStrategy
            sage: P.<x,y,z> = PolynomialRing(GF(19))
            sage: I = Ideal([P(0)])
            sage: strat = GroebnerStrategy(I)
            sage: strat == GroebnerStrategy(I)
            True
            sage: I = Ideal([x+1,y+z])
            sage: strat == GroebnerStrategy(I)
            False
        """
        try:
            lx = <GroebnerStrategy?>self
            rx = <GroebnerStrategy?>other
        except TypeError:
            return NotImplemented
        return richcmp(lx._ideal.gens(),
                       rx._ideal.gens(), op)

    def __reduce__(self):
        """
        EXAMPLES::

            sage: from sage.libs.singular.groebner_strategy import GroebnerStrategy
            sage: P.<x,y,z> = PolynomialRing(GF(32003))
            sage: I = Ideal([x + z, y + z])
            sage: strat = GroebnerStrategy(I)
            sage: loads(dumps(strat)) == strat
            True
        """
        return unpickle_GroebnerStrategy0, (self._ideal,)

    cpdef MPolynomial_libsingular normal_form(self, MPolynomial_libsingular p):
        """
        Compute the normal form of ``p`` with respect to the
        generators of this object.

        EXAMPLES::

            sage: from sage.libs.singular.groebner_strategy import GroebnerStrategy
            sage: P.<x,y,z> = PolynomialRing(QQ)
            sage: I = Ideal([x + z, y + z])
            sage: strat = GroebnerStrategy(I)
            sage: strat.normal_form(x*y) # indirect doctest
            z^2
            sage: strat.normal_form(x + 1)
            -z + 1

        TESTS::

            sage: from sage.libs.singular.groebner_strategy import GroebnerStrategy
            sage: P.<x,y,z> = PolynomialRing(QQ)
            sage: I = Ideal([P(0)])
            sage: strat = GroebnerStrategy(I)
            sage: strat.normal_form(x)
            x
            sage: strat.normal_form(P(0))
            0
        """
        if unlikely(p._parent is not self._parent):
            raise TypeError("parent(p) must be the same as this object's parent.")
        if unlikely(self._parent._ring != currRing):
            rChangeCurrRing(self._parent._ring)

        cdef int max_ind = 0
        cdef poly *_p = redNF(p_Copy(p._poly, self._parent._ring), max_ind, 0, self._strat)
        if likely(_p!=NULL):
            _p = redtailBba(_p, max_ind, self._strat)
        return new_MP(self._parent, _p)

cdef class NCGroebnerStrategy(SageObject):
    """
    A Wrapper for Singular's Groebner Strategy Object.

    This object provides functions for normal form computations and
    other functions for Groebner basis computation.

    ALGORITHM:

    Uses Singular via libSINGULAR
    """
    def __init__(self, L):
        """
        Create a new :class:`GroebnerStrategy` object for the
        generators of the ideal ``L``.

        INPUT:

        - ``L`` -- an ideal in a g-algebra

        EXAMPLES::

            sage: from sage.libs.singular.groebner_strategy import NCGroebnerStrategy

            sage: # needs sage.combinat
            sage: A.<x,y,z> = FreeAlgebra(QQ, 3)
            sage: H.<x,y,z> = A.g_algebra({y*x:x*y-z, z*x:x*z+2*x, z*y:y*z-2*y})
            sage: I = H.ideal([y^2, x^2, z^2-H.one()])
            sage: NCGroebnerStrategy(I) #random
            Groebner Strategy for ideal generated by 3 elements over Noncommutative Multivariate Polynomial Ring in x, y, z over Rational Field, nc-relations: {z*x: x*z + 2*x, z*y: y*z - 2*y, y*x: x*y - z}

        TESTS::

            sage: strat = NCGroebnerStrategy(None)
            Traceback (most recent call last):
            ...
            TypeError: First parameter must be an ideal in a g-algebra.

            sage: P.<x,y,z> = PolynomialRing(CC,order='neglex')
            sage: I = Ideal([x+z,y+z+1])
            sage: strat = NCGroebnerStrategy(I)
            Traceback (most recent call last):
            ...
            TypeError:  First parameter must be an ideal in a g-algebra.

        Check that tail reduction is applied too::

            sage: # needs sage.combinat
            sage: F = PolynomialRing(QQ,'t').fraction_field()
            sage: FA = FreeAlgebra(F, 6, 'x1,x2,x3,x4,x5,x6')
            sage: N = FA.g_algebra({FA.gen(j)*FA.gen(i):-FA.gen(i)*FA.gen(j) for i in range(5) for j in range(i+1,6)})
            sage: I = N.ideal([g^2 for g in N.gens()],side='twosided')
            sage: N.inject_variables()
            Defining x1, x2, x3, x4, x5, x6
            sage: I.reduce(x1*x2*x3 + x2^2*x4)
            x1*x2*x3
        """
        if not isinstance(L, NCPolynomialIdeal):
            raise TypeError("First parameter must be an ideal in a g-algebra.")

        if not isinstance(L.ring(), NCPolynomialRing_plural):
            raise TypeError("First parameter's ring must be a g-algebra.")

        self._ideal = L

        cdef NCPolynomialRing_plural R = <NCPolynomialRing_plural>L.ring()
        self._parent = R

        if not R.term_order().is_global():
            raise NotImplementedError("The local case is not implemented yet.")

        if not R.base_ring().is_field():
            raise NotImplementedError("Only coefficient fields are implemented so far.")

        if (R._ring != currRing):
            rChangeCurrRing(R._ring)

        cdef ideal *i = sage_ideal_to_singular_ideal(L)
        self._strat = new_skStrategy()

        self._strat.ak = id_RankFreeModule(i, R._ring)
        #- creating temp data structures
        initBuchMoraCrit(self._strat)
        self._strat.initEcart = initEcartBBA
        self._strat.enterS = enterSBba
        #- set S
        self._strat.sl = -1
        self._strat.noTailReduction = False
        #- init local data struct
        initS(i, NULL, self._strat)

        cdef int j
        if R.base_ring().is_field():
            for j in range(self._strat.sl,-1,-1):
                pNorm(self._strat.S[j])

        id_Delete(&i, R._ring)
        kTest(self._strat)

    def __dealloc__(self):
        """
        TESTS::

            sage: # needs sage.combinat
            sage: from sage.libs.singular.groebner_strategy import NCGroebnerStrategy
            sage: A.<x,y,z> = FreeAlgebra(QQ, 3)
            sage: H.<x,y,z> = A.g_algebra({y*x:x*y-z, z*x:x*z+2*x, z*y:y*z-2*y})
            sage: I = H.ideal([y^2, x^2, z^2-H.one()])
            sage: strat = NCGroebnerStrategy(I)
            sage: del strat   # indirect doctest
        """
        cdef ring *oldRing = NULL
        if self._strat:
            omfree(self._strat.sevS)
            omfree(self._strat.ecartS)
            omfree(self._strat.T)
            omfree(self._strat.sevT)
            omfree(self._strat.R)
            omfree(self._strat.S_2_R)
            omfree(self._strat.L)
            omfree(self._strat.B)
            omfree(self._strat.fromQ)
            id_Delete(&self._strat.Shdl, self._parent._ring)

            if self._parent._ring != currRing:
                oldRing = currRing
                rChangeCurrRing(self._parent._ring)
                delete_skStrategy(self._strat)
                rChangeCurrRing(oldRing)
            else:
                delete_skStrategy(self._strat)

    def _repr_(self):
        """
        TESTS::

            sage: # needs sage.combinat
            sage: from sage.libs.singular.groebner_strategy import NCGroebnerStrategy
            sage: A.<x,y,z> = FreeAlgebra(QQ, 3)
            sage: H.<x,y,z> = A.g_algebra({y*x:x*y-z, z*x:x*z+2*x, z*y:y*z-2*y})
            sage: I = H.ideal([y^2, x^2, z^2-H.one()])
            sage: strat = NCGroebnerStrategy(I)
            sage: strat # indirect doctest #random
            Groebner Strategy for ideal generated by 3 elements over Noncommutative Multivariate Polynomial Ring in x, y, z over Rational Field, nc-relations: {z*x: x*z + 2*x, z*y: y*z - 2*y, y*x: x*y - z}
        """
        return "Groebner Strategy for ideal generated by %d elements over %s" % (self._ideal.ngens(), self._parent)

    def ideal(self):
        """
        Return the ideal this strategy object is defined for.

        EXAMPLES::

            sage: # needs sage.combinat
            sage: from sage.libs.singular.groebner_strategy import NCGroebnerStrategy
            sage: A.<x,y,z> = FreeAlgebra(QQ, 3)
            sage: H.<x,y,z> = A.g_algebra({y*x:x*y-z, z*x:x*z+2*x, z*y:y*z-2*y})
            sage: I = H.ideal([y^2, x^2, z^2-H.one()])
            sage: strat = NCGroebnerStrategy(I)
            sage: strat.ideal() == I
            True
        """
        return self._ideal

    def ring(self):
        """
        Return the ring this strategy object is defined over.

        EXAMPLES::

            sage: # needs sage.combinat
            sage: from sage.libs.singular.groebner_strategy import NCGroebnerStrategy
            sage: A.<x,y,z> = FreeAlgebra(QQ, 3)
            sage: H.<x,y,z> = A.g_algebra({y*x:x*y-z, z*x:x*z+2*x, z*y:y*z-2*y})
            sage: I = H.ideal([y^2, x^2, z^2-H.one()])
            sage: strat = NCGroebnerStrategy(I)
            sage: strat.ring() is H
            True
        """
        return self._parent

    def __richcmp__(self, other, op):
        """
        EXAMPLES::

            sage: # needs sage.combinat
            sage: from sage.libs.singular.groebner_strategy import NCGroebnerStrategy
            sage: A.<x,y,z> = FreeAlgebra(QQ, 3)
            sage: H.<x,y,z> = A.g_algebra({y*x:x*y-z, z*x:x*z+2*x, z*y:y*z-2*y})
            sage: I = H.ideal([y^2, x^2, z^2-H.one()])
            sage: strat = NCGroebnerStrategy(I)
            sage: strat == NCGroebnerStrategy(I)
            True
            sage: I = H.ideal([y^2, x^2, z^2-H.one()], side='twosided')
            sage: strat == NCGroebnerStrategy(I)
            False
        """
        try:
            lx = <NCGroebnerStrategy?>self
            rx = <NCGroebnerStrategy?>other
        except TypeError:
            return NotImplemented
        return richcmp((lx._ideal.gens(), lx._ideal.side()),
                       (rx._ideal.gens(), rx._ideal.side()), op)

    def __reduce__(self):
        """
        EXAMPLES::

            sage: # needs sage.combinat
            sage: from sage.libs.singular.groebner_strategy import NCGroebnerStrategy
            sage: A.<x,y,z> = FreeAlgebra(QQ, 3)
            sage: H.<x,y,z> = A.g_algebra({y*x:x*y-z, z*x:x*z+2*x, z*y:y*z-2*y})
            sage: I = H.ideal([y^2, x^2, z^2-H.one()])
            sage: strat = NCGroebnerStrategy(I)
            sage: loads(dumps(strat)) == strat
            True
        """
        return unpickle_NCGroebnerStrategy0, (self._ideal,)

    cpdef NCPolynomial_plural normal_form(self, NCPolynomial_plural p):
        """
        Compute the normal form of ``p`` with respect to the
        generators of this object.

        EXAMPLES::

            sage: # needs sage.combinat
            sage: A.<x,y,z> = FreeAlgebra(QQ, 3)
            sage: H.<x,y,z> = A.g_algebra({y*x:x*y-z, z*x:x*z+2*x, z*y:y*z-2*y})
            sage: JL = H.ideal([x^3, y^3, z^3 - 4*z])
            sage: JT = H.ideal([x^3, y^3, z^3 - 4*z], side='twosided')
            sage: from sage.libs.singular.groebner_strategy import NCGroebnerStrategy
            sage: SL = NCGroebnerStrategy(JL.std())
            sage: ST = NCGroebnerStrategy(JT.std())
            sage: SL.normal_form(x*y^2)
            x*y^2
            sage: ST.normal_form(x*y^2)
            y*z
        """
        if unlikely(p._parent is not self._parent):
            raise TypeError("parent(p) must be the same as this object's parent.")
        if unlikely(self._parent._ring != currRing):
            rChangeCurrRing(self._parent._ring)

        cdef int max_ind = 0
        cdef poly *_p = redNF(p_Copy(p._poly, self._parent._ring), max_ind, 0, self._strat)
        if likely(_p!=NULL):
            _p = redtailBba(_p, max_ind, self._strat)
        return new_NCP(self._parent, _p)


def unpickle_NCGroebnerStrategy0(I):
    """
    EXAMPLES::

        sage: # needs sage.combinat
        sage: from sage.libs.singular.groebner_strategy import NCGroebnerStrategy
        sage: A.<x,y,z> = FreeAlgebra(QQ, 3)
        sage: H.<x,y,z> = A.g_algebra({y*x:x*y-z, z*x:x*z+2*x, z*y:y*z-2*y})
        sage: I = H.ideal([y^2, x^2, z^2-H.one()])
        sage: strat = NCGroebnerStrategy(I)
        sage: loads(dumps(strat)) == strat   # indirect doctest
        True
    """
    return NCGroebnerStrategy(I)


def unpickle_GroebnerStrategy0(I):
    """
    EXAMPLES::

        sage: # needs sage.rings.finite_rings
        sage: from sage.libs.singular.groebner_strategy import GroebnerStrategy
        sage: P.<x,y,z> = PolynomialRing(GF(32003))
        sage: I = Ideal([x + z, y + z])
        sage: strat = GroebnerStrategy(I)
        sage: loads(dumps(strat)) == strat # indirect doctest
        True
    """
    return GroebnerStrategy(I)



# sage_setup: distribution = sagemath-singular
from sage.libs.singular.decl cimport skStrategy, ring

from sage.rings.polynomial.multi_polynomial_libsingular cimport \
    MPolynomialRing_libsingular, MPolynomial_libsingular
from sage.structure.sage_object cimport SageObject
from sage.rings.polynomial.plural cimport NCPolynomialRing_plural, NCPolynomial_plural

cdef class GroebnerStrategy(SageObject):
    cdef skStrategy *_strat
    cdef ring *_parent_ring
    cdef MPolynomialRing_libsingular _parent
    cdef object _ideal

    cpdef MPolynomial_libsingular normal_form(self, MPolynomial_libsingular p)

cdef class NCGroebnerStrategy(SageObject):
    cdef skStrategy *_strat
    cdef NCPolynomialRing_plural _parent
    cdef object _ideal

    cpdef NCPolynomial_plural normal_form(self, NCPolynomial_plural p)



# sage_setup: distribution = sagemath-singular
"""
Direct Access to Singular's Functions via libSingular

AUTHORS:

- Michael Brickenstein (2009-07): initial implementation, overall design
- Martin Albrecht (2009-07): clean up, enhancements, etc.
"""
#*****************************************************************************
#       Copyright (C) 2009 Michael Brickenstein <brickenstein@mfo.de>
#       Copyright (C) 2009 Martin Albrecht <M.R.Albrecht@rhul.ac.uk>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
#*****************************************************************************

from sage.structure.sage_object cimport SageObject
from sage.libs.singular.decl cimport leftv, idhdl, syStrategy, matrix, poly, ideal, intvec
from sage.libs.singular.decl cimport ring as singular_ring
from sage.rings.polynomial.multi_polynomial_libsingular cimport MPolynomialRing_libsingular, MPolynomial_libsingular

cdef new_sage_polynomial(ring,  poly *p)
cdef poly* access_singular_poly(p) except <poly*> -1
cdef singular_ring* access_singular_ring(r) except <singular_ring*> -1

cdef class RingWrap:
    cdef singular_ring *_ring

cdef class Resolution:
    cdef syStrategy *_resolution
    cdef object base_ring

cdef class Converter(SageObject):
    cdef leftv *args
    cdef object _sage_ring
    cdef singular_ring* _singular_ring
    cdef leftv* pop_front(self) except NULL
    cdef leftv * _append_leftv(self, leftv *v) noexcept
    cdef leftv * _append(self, void* data, int res_type) noexcept
    cdef leftv * append_polynomial(self, p) except NULL
    cdef leftv * append_ideal(self,  i) except NULL
    cdef leftv * append_number(self, n) except NULL
    cdef leftv * append_int(self, n) except NULL
    cdef leftv * append_str(self, n) except NULL
    cdef leftv * append_intmat(self, a) except NULL
    cdef leftv * append_resolution(self, Resolution resolution) except NULL
    cdef leftv * append_vector(self, v) except NULL
    cdef leftv * append_intvec(self, v) except NULL
    cdef leftv * append_list(self, l) except NULL
    cdef leftv * append_matrix(self, a) except NULL
    cdef leftv * append_ring(self, r) except NULL
    cdef leftv * append_module(self, m) except NULL
    cdef to_sage_integer_matrix(self, intvec *mat)
    cdef object to_sage_module_element_sequence_destructive(self, ideal *i)
    cdef to_sage_vector_destructive(self, poly *p, free_module=?)
    cdef to_sage_matrix(self, matrix* mat)
    cdef to_python(self, leftv* to_convert)

cdef class BaseCallHandler:
    cdef leftv* handle_call(self, Converter argument_list, singular_ring *_ring=?) noexcept
    cdef bint free_res(self) noexcept

cdef class LibraryCallHandler(BaseCallHandler):
    cdef idhdl * proc_idhdl

cdef class KernelCallHandler(BaseCallHandler):
    cdef long arity
    cdef long cmd_n

cdef class SingularFunction(SageObject):
    cdef object _name
    cdef MPolynomialRing_libsingular _ring
    cdef BaseCallHandler call_handler

    cdef BaseCallHandler get_call_handler(self)
    cdef bint function_exists(self) noexcept
    cdef common_ring(self, tuple args, ring=?)

cdef class SingularLibraryFunction(SingularFunction):
    pass

cdef class SingularKernelFunction(SingularFunction):
    pass

# the most direct function call interface
cdef call_function(SingularFunction self, tuple args, object R, bint signal_handler=?, object attributes=?)



# sage_setup: distribution = sagemath-singular
# sage.doctest: needs sage.combinat sage.groups
"""
The Fusion Ring of the Drinfeld Double of a Finite Group
"""

# ****************************************************************************
#  Copyright (C) 2023 Wenqi Li
#                     Daniel Bump <bump at match.stanford.edu>
#                     Travis Scrimshaw <tcscrims at gmail.com>
#                     Guillermo Aboumrad <gh_willieab>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.categories.algebras_with_basis import AlgebrasWithBasis
from sage.combinat.free_module import CombinatorialFreeModule
from sage.rings.integer_ring import ZZ
from sage.misc.misc import inject_variable
from sage.misc.cachefunc import cached_method
from sage.sets.set import Set
from sage.rings.number_field.number_field import CyclotomicField
from sage.matrix.constructor import matrix


class FusionDouble(CombinatorialFreeModule):
    r"""
    The fusion ring corresponding to the Drinfeld double of a finite group.

    This is the fusion ring of the modular tensor category of modules
    over the Drinfeld double of a finite group. Usage is similar
    to :class:`FusionRing`; we refer the reader to that class for more
    information.

    INPUT:

    - ``G`` -- a finite group
    - ``prefix`` -- (default: ``'s'``) a prefix for the names of simple objects
    - ``inject_varables`` -- (optional) set to ``True`` to create variables
      for the simple objects

    REFERENCES:

    - [BaKi2001]_ Chapter 3
    - [Mas1995]_
    - [CHW2015]_
    - [Goff1999]_

    EXAMPLES::

        sage: G = DihedralGroup(5)
        sage: H = FusionDouble(G, inject_variables=True)
        sage: H.basis()
        Finite family {0: s0, 1: s1, 2: s2, 3: s3, 4: s4, 5: s5, 6: s6, 7: s7, 8: s8,
                       9: s9, 10: s10, 11: s11, 12: s12, 13: s13, 14: s14, 15: s15}
        sage: for x in H.basis():
        ....:     print ("%s : %s"%(x,x^2))
        ....:
        s0 : s1
        s1 : s1
        s2 : s0 + s1 + s3
        s3 : s0 + s1 + s2
        s4 : s0 + s1 + s13
        s5 : s0 + s1 + s10
        s6 : s0 + s1 + s12
        s7 : s0 + s1 + s9
        s8 : s0 + s1 + s11
        s9 : s0 + s1 + s7
        s10 : s0 + s1 + s5
        s11 : s0 + s1 + s8
        s12 : s0 + s1 + s6
        s13 : s0 + s1 + s4
        s14 : s1 + s2 + s3 + s4 + s5 + s6 + s7 + s8 + s9 + s10 + s11 + s12 + s13
        s15 : s1 + s2 + s3 + s4 + s5 + s6 + s7 + s8 + s9 + s10 + s11 + s12 + s13
        sage: s4*s5
        s2 + s9
        sage: s4.ribbon()
        -zeta5^3 - zeta5^2 - zeta5 - 1
        sage: s5.ribbon()
        1
        sage: s8.ribbon()
        zeta5^3

    If the fusion double is multiplicity-free, meaning that the fusion
    coefficients `N_k^{ij}` are bounded by `1`, then the F-matrix may be
    computed, by solving the pentagon and hexagon relations as described
    in [Bond2007]_ and [Ab2022]_, just as for :class:`FusionRing`.
    There is a caveat here, since even if the fusion rules are multiplicity-free,
    if there are too many F-matrix values to compute, even if many of them are
    zero, in the current implementation singular cannot create enough variables.
    At least, this code can compute the F-matrix for the Fusion Double of the
    symmetric group `S_3`, duplicating the result of [CHW2015]_.

    ::

        sage: G1 = SymmetricGroup(3)
        sage: H1 = FusionDouble(G1, prefix='u', inject_variables=True)
        sage: F = H1.get_fmatrix()

    The above commands create the F-matrix. You can compute all of the
    F-matrices with the command::

        sage: H1.find_orthogonal_solution()  # not tested (10-15 minutes)

    Individual F-matrices may be computed thus::

        sage: F.fmatrix(u3, u3, u3, u4)  # not tested

    See :class:`FMatrix` for more information.

    Unfortunately beyond `S_3` the number of simple objects is seemingly
    impractical. Although the :class:`FusionDouble` class and its methods
    work well for groups of moderate size, the :class:`FMatrix` may not be
    computable. For the dihedral group of order 8, there are already 22
    simple objects, and the F-matrix seems out of reach. The actual limitation
    is that singular will not create a polynomial ring in more than
    `2^{15}-1 = 32767` symbols, and there are more than this many F-matrix
    values to be computed for the dihedral group of order 8, so in the
    current implementation, this FusionRing is out of reach.

    It is an open problem to classify the finite groups whose fusion doubles
    are multiplicity-free. Abelian groups, dihedral groups, dicyclic groups,
    and all groups of order 16 are multiplicity-free.  On the other hand, for
    groups of order 32, some are multiplicity-free and others are not.
    These can all be constructed using :class:`SmallPermutationGroup`.

    EXAMPLES::

        sage: G = SmallPermutationGroup(16,9)
        sage: F = FusionDouble(G, prefix='b', inject_variables=True)
        sage: b13^2 # long time (4s)
        b0 + b3 + b37
    """
    @staticmethod
    def __classcall_private__(cls, G, prefix='s', inject_variables=False):
        """
        Normalize input to ensure a unique representation.

        EXAMPLES::

            sage: H1 = FusionDouble(DihedralGroup(6), inject_variables=True)
            sage: H2 = FusionDouble(DihedralGroup(6), prefix='s')
            sage: H1 is H2
            True
        """
        F = super().__classcall__(cls, G=G, prefix=prefix)
        if inject_variables:
            F.inject_variables()
        return F

    def __init__(self, G, prefix='s') -> None:
        """
        EXAMPLES::

            sage: H = FusionDouble(DihedralGroup(6))
            sage: TestSuite(H).run()
            sage: H = FusionDouble(DihedralGroup(7))
            sage: TestSuite(H).run()  # long time

            sage: F = FusionDouble(CyclicPermutationGroup(2))
            sage: [F._repr_term(t) for t in F._names]
            ['s0', 's1', 's2', 's3']
            sage: F = FusionDouble(CyclicPermutationGroup(2))
            sage: [F._latex_term(t) for t in F._names]
            ['s_{0}', 's_{1}', 's_{2}', 's_{3}']

            sage: FusionDouble(SymmetricGroup(4)).get_order()
            [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20]
        """
        self._G = G
        self._names = {}
        self._elt = {}
        self._chi = {}
        self._unit_index = None  # index of the unit element
        count = ZZ.zero()
        for g in sorted(G.conjugacy_classes_representatives(), key=str):
            for chi in sorted(G.centralizer(g).irreducible_characters(), key=lambda chi:str(chi.values())):
                # NOTE: the trivial char is not necessarily the first one
                self._names[count] = "%s%s" % (prefix, count)
                self._elt[count] = g
                self._chi[count] = chi
                if self._unit_index is None and all(v == 1 for v in chi):
                    self._unit_index = count
                count += ZZ.one()
        self._cyclotomic_order = G.exponent()
        self._basecoer = None
        self._fusion_labels = None
        self._field = None
        cat = AlgebrasWithBasis(ZZ)
        CombinatorialFreeModule.__init__(self, ZZ, list(self._names),
                                         prefix=prefix, bracket=False, category=cat)

    def _repr_(self) -> str:
        """
        EXAMPLES::

            sage: FusionDouble(SymmetricGroup(3))
            The Fusion Ring of the Drinfeld Double of Symmetric group of
                order 3! as a permutation group
        """
        return "The Fusion Ring of the Drinfeld Double of %s" % self._G

    def inject_variables(self):
        """
        Create variables for the simple objects in the global name space.

        EXAMPLES::

            sage: F = FusionDouble(DiCyclicGroup(3), prefix='d')
            sage: F.inject_variables()
            sage: d0 + d1 + d5
            d0 + d1 + d5
        """
        for i, name in self._names.items():
            inject_variable(name, self.monomial(i))

    @cached_method
    def _char_cache(self, i, g):
        """
        Return ``self._chi[i](g)``, cached here for speed.

        TESTS::

            sage: D = FusionDouble(SymmetricGroup(4))

            sage: all(D._char_cache(b.support_of_term(), b.g()) == b.char()(b.g())
            ....:     for b in D.basis())
            True
        """
        return self._chi[i](g)

    @cached_method
    def s_ij(self, i, j, unitary=False, base_coercion=True):
        r"""
        Return the element of the `S`-matrix of this fusion ring
        corresponding to the given elements.

        Without the unitary option set true, this is the unnormalized `S`-matrix
        entry, denoted `\tilde{s}_{ij}`, in [BaKi2001]_ Chapter 3. The
        normalized `S`-matrix entries are denoted `s_{ij}`.

        INPUT:

        - ``i``, ``j``, -- a pair of basis elements
        - ``unitary`` -- boolean (default: ``False``); set to ``True`` to
          obtain the unitary `S`-matrix

        EXAMPLES::

            sage: D = FusionDouble(SymmetricGroup(3), prefix='t', inject_variables=True)
            sage: [D.s_ij(t2, x) for x in D.basis()]
            [2, 2, 4, 0, 0, -2, -2, -2]
            sage: [D.s_ij(t2, x, unitary=True) for x in D.basis()]
            [1/3, 1/3, 2/3, 0, 0, -1/3, -1/3, -1/3]
        """
        sum_val = ZZ.zero()
        G = self._G
        i, = list(i._monomial_coefficients)
        j, = list(j._monomial_coefficients)
        a = self._elt[i]
        b = self._elt[j]
        for g in G:
            gi = g.inverse()
            conj = g * b * gi
            if a * conj == conj * a:
                sum_val += self._char_cache(i, conj) * self._char_cache(j, gi * a * g)
        if unitary:
            coef = 1 / (G.centralizer(a).order() * G.centralizer(b).order())
        else:
            coef = G.order() / (G.centralizer(a).order() * G.centralizer(b).order())
        ret = coef * sum_val
        if (not base_coercion) or (self._basecoer is None):
            return ret
        return self._basecoer(ret)

    def s_ijconj(self, i, j, unitary=False, base_coercion=True):
        r"""
        Return the conjugate of the element of the `S`-matrix given by
        ``self.s_ij(elt_i, elt_j, base_coercion=base_coercion)``.

        .. SEEALSO::

            :meth:`s_ij`

        EXAMPLES::

            sage: P = FusionDouble(CyclicPermutationGroup(3),prefix='p',inject_variables=True)
            sage: P.s_ij(p1,p3)
            1
            sage: P.s_ijconj(p1,p3)
            1
        """
        return self.s_ij(i, j, unitary=unitary, base_coercion=base_coercion).conjugate()

    def s_matrix(self, unitary=False, base_coercion=True):
        r"""
        Return the `S`-matrix of this fusion ring.

        OPTIONAL:

        - ``unitary`` -- boolean (default: ``False``); set to ``True`` to
          obtain the unitary `S`-matrix

        Without the ``unitary`` parameter, this is the matrix denoted
        `\widetilde{s}` in [BaKi2001]_.

        EXAMPLES::

            sage: FusionDouble(SymmetricGroup(3)).s_matrix()
            [ 1  1  2 -3 -3  2  2  2]
            [ 1  1  2  3  3  2  2  2]
            [ 2  2  4  0  0 -2 -2 -2]
            [-3  3  0  3 -3  0  0  0]
            [-3  3  0 -3  3  0  0  0]
            [ 2  2 -2  0  0 -2 -2  4]
            [ 2  2 -2  0  0 -2  4 -2]
            [ 2  2 -2  0  0  4 -2 -2]
            sage: FusionDouble(SymmetricGroup(3)).s_matrix(unitary=True)
            [ 1/6  1/6  1/3 -1/2 -1/2  1/3  1/3  1/3]
            [ 1/6  1/6  1/3  1/2  1/2  1/3  1/3  1/3]
            [ 1/3  1/3  2/3    0    0 -1/3 -1/3 -1/3]
            [-1/2  1/2    0  1/2 -1/2    0    0    0]
            [-1/2  1/2    0 -1/2  1/2    0    0    0]
            [ 1/3  1/3 -1/3    0    0 -1/3 -1/3  2/3]
            [ 1/3  1/3 -1/3    0    0 -1/3  2/3 -1/3]
            [ 1/3  1/3 -1/3    0    0  2/3 -1/3 -1/3]
        """
        b = self.basis()
        return matrix([[self.s_ij(b[x], b[y], unitary=unitary,
                                  base_coercion=base_coercion)
                        for x in self.get_order()] for y in self.get_order()])

    @cached_method
    def N_ijk(self, i, j, k):
        r"""
        The symmetric invariant of three simple objects.

        This is the dimension of

        .. MATH::

           Hom(i \otimes j \otimes k, s_0),

        where `s_0` is the unit element (assuming ``prefix='s'``).
        Method of computation is through the Verlinde formula,
        deducing the values from the known values of the `S`-matrix.

        EXAMPLES::

            sage: A = FusionDouble(AlternatingGroup(4),prefix='a',inject_variables=True)
            sage: [A.N_ijk(a10,a11,x) for x in A.basis()]
            [1, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0]

        TESTS::

            sage: F = FusionDouble(SymmetricGroup(4))
            sage: from itertools import product
            sage: B = list(F.basis())
            sage: all(F.N_ijk(i,j,k).parent() is ZZ
            ....:     for i, j, k in product(B[::6], repeat=3))
            True
        """
        sz = self.one()
        return ZZ(sum(self.s_ij(i, r, unitary=True) * self.s_ij(j, r, unitary=True)
                      * self.s_ij(k, r, unitary=True) / self.s_ij(sz, r, unitary=True)
                      for r in self.basis()))

    @cached_method
    def Nk_ij(self, i, j, k, use_characters=False):
        r"""
        Return the fusion coefficient `N^k_{ij}`.

        INPUT:

        - ``i``, ``j``, ``k`` -- basis elements
        - ``use_characters`` -- boolean (default: ``False``); see the algorithm
          description below

        ALGORITHM:

        If ``use_characters=False``, then this is computed using
        the Verlinde formula:

        .. MATH::

            N^k_{ij} = \sum_l \frac{s(i, \ell)\, s(j, \ell)\,
            \overline{s(k, \ell)}}{s(I, \ell)}.

        Otherwise we use a character theoretic method to compute the fusion
        coefficient `N_{ij}^k` as follows. Each simple object, for example
        `i` corresponds to a conjugacy class `\mathcal{C}_i` of the underlying
        group `G`, and an irreducible character `\chi_i` of the centralizer
        `C(g_i)` of a fixed representative `g_i` of `\mathcal{C}_i`. In addition
        to the fixed representative `g_k` of the class `\mathcal{C}_i`
        and `\mathcal{C}_j`, the formula will make use of variable elements
        `h_i` and `h_j` that are subject to the condition `h_i h_j = g_k`.
        See [GoMa2010]_ equation (7).

        .. MATH::

            \frac{|\mathcal{C}_k|}{|G|}
            \sum_{\substack{h_i\in\mathcal{C}_i \\ h_j\in\mathcal{C}_j \\ h_ih_j=g_k}}
            \lvert C(h_i)\cap C(h_j) \rvert \,
            \langle \chi_i^{(h_i)} \chi_j^{(h_j)}, \chi_k \rangle_{C(h_i)\cap C(h_j)},

        where `\chi_i^{(h_i)}` is the character `\chi_i` of `C(g_i)`
        conjugated to a character of `C(h_i)`, when `h_i` is a conjugate
        of the fixed representative `g_i`. More exactly, there exists `r_i`
        such that `r_i g_i r_i^{-1} = h_i`, and then `\chi_i^{(h_i)}(x) =
        \chi_i(r_i^{-1}xr_i)`, and this definition does not depend on the
        choice of `r_i`.

        .. NOTE::

            This should be functionally equivalent, and testing shows
            that it is, but it is slower.

        EXAMPLES::

            sage: A = FusionDouble(AlternatingGroup(4),prefix='aa',inject_variables=True)
            sage: [A.Nk_ij(aa8,aa10,x) for x in A.basis()]
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 2, 1]

            sage: B = FusionDouble(CyclicPermutationGroup(2))
            sage: all(B.Nk_ij(x,y,z,use_characters=True) == B.Nk_ij(x,y,z)
            ....:     for x in B.basis() for y in B.basis() for z in B.basis())
            True
        """
        if not use_characters:
            return self.N_ijk(i, j, self.dual(k))

        G = self._G
        I = G.conjugacy_class(i.g())
        J = G.conjugacy_class(j.g())
        IJ = {I_elem * J_elem for I_elem in I for J_elem in J}
        if k.g() not in IJ:
            return ZZ.zero()

        K = G.conjugacy_class(k.g())
        CI = G.centralizer(i.g())
        CJ = G.centralizer(j.g())
        CK = G.centralizer(k.g())

        c = K.cardinality() / G.order()
        summands = [(I_elem, J_elem) for I_elem in I for J_elem in J if I_elem * J_elem == k.g()]
        res = ZZ.zero()
        ichar = i.char()
        jchar = j.char()
        kchar = k.char()
        for p in summands:
            I_elem, J_elem = p
            for g in G:
                ginv = g.inverse()
                if ginv * i.g() * g == I_elem:
                    i_twist = g
                if ginv * j.g() * g == J_elem:
                    j_twist = g
            A = Set(i_twist.inverse() * zi * i_twist for zi in CI)
            B = Set(j_twist.inverse() * zj * j_twist for zj in CJ)
            inner_summands = A.intersection(B).intersection(Set(CK))
            i_twist_inv = i_twist.inverse()
            j_twist_inv = j_twist.inverse()
            res += sum(ichar(i_twist * x * i_twist_inv) * jchar(j_twist * x * j_twist_inv) * kchar(x).conjugate()
                       for x in inner_summands)
        return c * res

    @cached_method
    def field(self):
        """
        Return a cyclotomic field large enough to contain the values
        of R-matrices and twists that can arise for this fusion ring.

        EXAMPLES::

            sage: FusionDouble(SymmetricGroup(3)).field()
            Cyclotomic Field of order 24 and degree 8
        """
        return CyclotomicField(4 * self._cyclotomic_order)

    def fvars_field(self):
        r"""
        Return a field containing the ``CyclotomicField`` computed by
        :meth:`field` as well as all the F-symbols of the associated
        ``FMatrix`` factory object.

        This method is only available if ``self`` is multiplicity-free.

        EXAMPLES::

            sage: FusionDouble(SymmetricGroup(3)).fvars_field()
            Cyclotomic Field of order 24 and degree 8
        """
        if self.is_multiplicity_free(verbose=False):
            return self.get_fmatrix().field()
        raise NotImplementedError("method is only available for multiplicity free fusion rings")

    def root_of_unity(self, r, base_coercion=True):
        r"""
        Return `e^{i\pi r}` as an element of ``self.field()`` if possible.

        INPUT:

        - ``r`` -- a rational number

        EXAMPLES::

            sage: H = FusionDouble(DihedralGroup(6))
            sage: H.field()
            Cyclotomic Field of order 24 and degree 8
            sage: for n in [1..7]:
            ....:     try:
            ....:         print (n,H.root_of_unity(2/n))
            ....:     except ValueError as err:
            ....:         print (n,err)
            ....:
            1 1
            2 -1
            3 zeta24^4 - 1
            4 zeta24^6
            5 not a root of unity in the field
            6 zeta24^4
            7 not a root of unity in the field
        """
        n = 2 * r * self._cyclotomic_order
        if n not in ZZ:
            raise ValueError("not a root of unity in the field")
        ret = self.field().gen() ** n
        if (not base_coercion) or (self._basecoer is None):
            return ret
        return self._basecoer(ret)

    @cached_method
    def r_matrix(self, i, j, k, base_coercion=True):
        r"""
        Return the R-matrix entry corresponding to the subobject ``k``
        in the tensor product of ``i`` with ``j``. This method is only
        correct if the fusion coefficient ``N_{ij}^k\leq 1``. See the
        :class:`FusionRing` method for more information, including
        the reason for this caveat, and the algorithm.

        EXAMPLES::

            sage: C = FusionDouble(SymmetricGroup(3),prefix='c',inject_variables=True)
            sage: c4*c5
            c3 + c4
            sage: [C.r_matrix(c4,c5,k) for k in [c3,c4]]
            [-zeta24^6 + zeta24^2, -zeta24^4]
            sage: c6^2
            c0 + c1 + c6
            sage: [C.r_matrix(c6,c6,k) for k in [c0,c1,c6]]
            [-1, 1, 1]
        """
        if self.Nk_ij(i, j, k) == 0:
            return self.field().zero() if (not base_coercion) or (self._basecoer is None) else self.fvars_field().zero()

        if i != j:
            ret = self.root_of_unity((k.twist() - i.twist() - j.twist()) / 2)
        else:
            i0 = self.one()
            B = self.basis()
            ret = sum(y.ribbon()**2 / (i.ribbon() * x.ribbon()**2)
                      * self.s_ij(i0, y) * self.s_ij(i, z) * self.s_ijconj(x, z)
                      * self.s_ijconj(k, x) * self.s_ijconj(y, z) / self.s_ij(i0, z)
                      for x in B for y in B for z in B) / (self.total_q_order()**4)

        if (not base_coercion) or (self._basecoer is None):
            return ret
        return self._basecoer(ret)

    def global_q_dimension(self, base_coercion=True):
        r"""
        Return the global quantum dimension, which is the sum of the squares of the
        quantum dimensions of the simple objects.

        For the Drinfeld double, it is the square of the order of the underlying quantum group.

        EXAMPLES::

            sage: G = SymmetricGroup(4)
            sage: H = FusionDouble(G)
            sage: H.global_q_dimension()
            576
            sage: sum(x.q_dimension()^2 for x in H.basis())
            576
        """
        ret = self._G.order() ** 2
        if (not base_coercion) or (self._basecoer is None):
            return ret
        return self._basecoer(ret)

    def total_q_order(self, base_coercion=True):
        r"""
        Return the positive square root of :meth:`self.global_q_dimension()
        <global_q_dimension>` as an element of :meth:`self.field() <field>`.

        For the Drinfeld double of a finite group `G`, this equals the
        cardinality of `G`. This is also equal to `\sum d_i^2 \theta_i^{\pm 1}`,
        where `i` runs through the simple objects, `d_i` is the quantum
        dimension, and `\theta_i` is the twist. This sum with `\theta_i` is
        denoted `p_-` in [BaKi2001]_ Chapter 3.

        EXAMPLES::

            sage: FusionDouble(DihedralGroup(7)).total_q_order()
            14
        """
        ret = self._G.order()
        if (not base_coercion) or (self._basecoer is None):
            return ret
        return self._basecoer(ret)

    D_minus = D_plus = total_q_order

    def is_multiplicity_free(self, verbose=False) -> bool:
        """
        Return ``True`` if all fusion coefficients are at most 1.

        EXAMPLES::

            sage: FusionDouble(SymmetricGroup(3)).is_multiplicity_free()
            True
            sage: FusionDouble(SymmetricGroup(4)).is_multiplicity_free()
            False

            sage: FusionDouble(SymmetricGroup(3)).is_multiplicity_free(True)
            Checking multiplicity freeness
            True
            sage: FusionDouble(SymmetricGroup(4)).is_multiplicity_free(True)
            Checking multiplicity freeness
            N(s2,s13,s13) = 2
            False
        """
        if verbose:
            print("Checking multiplicity freeness")
            from itertools import product
            for (i, j, k) in product(self.basis(), repeat=3):
                if self.N_ijk(i, j, k) > 1:
                    print("N(%s,%s,%s) = %s" % (i, j, k, self.N_ijk(i, j, k)))
                    return False
            return True

        return all(self.N_ijk(i, j, k) <= 1 for i in self.basis()
                   for j in self.basis() for k in self.basis())

    @cached_method
    def one_basis(self):
        r"""
        The unit element of the ring, which is the first basis element.

        EXAMPLES::

            sage: FusionDouble(CyclicPermutationGroup(2), prefix='h').one()
            h1
        """
        return self._unit_index

    @cached_method
    def dual(self, i):
        r"""
        Return the dual object ``i^\ast`` to ``i``.

        The dual is also available as an element method of ``i``.

        EXAMPLES::

            sage: K = FusionDouble(CyclicPermutationGroup(3),prefix='k')
            sage: [(x,K.dual(x)) for x in K.basis()]
            [(k0, k2),
             (k1, k1),
             (k2, k0),
             (k3, k8),
             (k4, k7),
             (k5, k6),
             (k6, k5),
             (k7, k4),
             (k8, k3)]
            sage: all(K.dual(x)==x.dual() for x in K.basis())
            True
        """
        sz = self.one()
        for j in self.basis():
            if self.N_ijk(i, j, sz) > 0:
                return j

    def product_on_basis(self, a, b):
        """
        Return the product of two basis elements corresponding to keys `a` and `b`.

        INPUT:

        - ``a``, ``b`` -- keys for the dictionary ``self._names`` representing
          simple objects

        EXAMPLES::

            sage: Q = FusionDouble(SymmetricGroup(3),prefix='q',inject_variables=True)
            sage: q3*q4
            q0 + q2 + q5 + q6 + q7
            sage: Q._names
            {0: 'q0', 1: 'q1', 2: 'q2', 3: 'q3', 4: 'q4', 5: 'q5', 6: 'q6', 7: 'q7'}
            sage: Q.product_on_basis(3,4)
            q0 + q2 + q5 + q6 + q7
        """
        d = {k.support_of_term(): val for k in self.basis()
             if (val := self.N_ijk(self.monomial(a), self.monomial(b),
                                   self.dual(k)))}
        return self._from_dict(d, remove_zeros=False)

    def group(self):
        """
        Return the underlying group.

        EXAMPLES::

            sage: FusionDouble(DiCyclicGroup(4)).group()
            Dicyclic group of order 16 as a permutation group
        """
        return self._G

    def get_fmatrix(self, *args, **kwargs):
        r"""
        Construct an :class:`FMatrix` factory to solve the pentagon and
        hexagon relations and organize the resulting F-symbols.

        EXAMPLES::

            sage: f = FusionDouble(SymmetricGroup(3)).get_fmatrix(); f
            F-Matrix factory for The Fusion Ring of the Drinfeld Double of
                Symmetric group of order 3! as a permutation group
        """
        if not hasattr(self, 'fmats') or kwargs.get('new', False):
            kwargs.pop('new', None)
            from sage.algebras.fusion_rings.f_matrix import FMatrix
            self.fmats = FMatrix(self, *args, **kwargs)
        return self.fmats

    class Element(CombinatorialFreeModule.Element):
        def is_simple_object(self) -> bool:
            r"""
            Determine whether ``self`` is a simple object (basis element) of the fusion ring.

            EXAMPLES::

                sage: H = FusionDouble(CyclicPermutationGroup(2), prefix='g', inject_variables=True)
                sage: [x.is_simple_object() for x in [g0, g1, g0+g1]]
                [True, True, False]
            """
            return len(self._monomial_coefficients) == 1

        def g(self):
            r"""
            The data determining a simple object consists of a conjugacy
            class representative `g` and an irreducible character `\chi` of
            the centralizer of `g`.

            This returns the conjugacy class representative of the underlying
            group corresponding to a simple object. See also :meth:`char`.

            EXAMPLES::

                sage: G = QuaternionGroup()
                sage: H = FusionDouble(G, prefix='e', inject_variables=True)
                sage: e10.g()
                (1,3)(2,4)(5,7)(6,8)
                sage: e10.char()
                Character of Subgroup generated by [(1,2,3,4)(5,6,7,8), (1,5,3,7)(2,8,4,6)]
                    of (Quaternion group of order 8 as a permutation group)
            """
            return self.parent()._elt[self.support_of_term()]

        def char(self):
            r"""
            Return the character `\chi` corresponding to ``self``.

            The data determining a simple object consists of a conjugacy
            class representative `g` and an irreducible character `\chi` of
            the centralizer of `g`.

            .. SEEALSO:: :meth:`g`

            EXAMPLES::

                sage: G = DihedralGroup(5)
                sage: H = FusionDouble(G, prefix='f', inject_variables=True)
                sage: f10.g()
                (1,3,5,2,4)
                sage: f10.char()
                Character of Subgroup generated by [(1,3,5,2,4), (1,2,3,4,5)] of
                    (Dihedral group of order 10 as a permutation group)
            """
            return self.parent()._chi[self.support_of_term()]

        def ribbon(self, base_coercion=True):
            """
            Return the twist or ribbon of the simple object.

            EXAMPLES::

                sage: H = FusionDouble(CyclicPermutationGroup(3))
                sage: [i.ribbon() for i in H.basis()]
                [1, 1, 1, -zeta3 - 1, 1, zeta3, zeta3, 1, -zeta3 - 1]
            """
            i = self.support_of_term()
            P = self.parent()
            ret = P._char_cache(i, self.g()) / P._char_cache(i, P._G.one())
            if (not base_coercion) or (P._basecoer is None):
                return ret
            return self.parent()._basecoer(ret)

        def twist(self, reduced=True):
            r"""
            Return a rational number `h` such that `\theta = e^{i \pi h}`
            is the twist of ``self``.

            The quantity `e^{i \pi h}` is also available using :meth:`ribbon`.

            This method is only available for simple objects.

            EXAMPLES::

                sage: Q = FusionDouble(CyclicPermutationGroup(3))
                sage: [x.twist() for x in Q.basis()]
                [0, 0, 0, 4/3, 0, 2/3, 2/3, 0, 4/3]
                sage: [x.ribbon() for x in Q.basis()]
                [1, 1, 1, -zeta3 - 1, 1, zeta3, zeta3, 1, -zeta3 - 1]

            TESTS::

                sage: H = FusionDouble(AlternatingGroup(4))
                sage: sum(H.basis()).twist()
                Traceback (most recent call last):
                ...
                ValueError: quantum twist is only available for simple objects
            """
            if not self.is_simple_object():
                raise ValueError("quantum twist is only available for simple objects")
            P = self.parent()
            zeta = P.field().gen()
            rib = self.ribbon()
            norm = 2 * P._cyclotomic_order
            for k in range(4 * P._cyclotomic_order):
                if zeta ** k == rib:
                    return k / norm

        def dual(self):
            """
            Return the dual of ``self``.

            This method is only available for simple objects.

            EXAMPLES::

                sage: G = CyclicPermutationGroup(4)
                sage: H = FusionDouble(G, prefix='j')
                sage: [x for x in H.basis() if x == x.dual()]
                [j0, j2, j8, j10]

            TESTS::

                sage: H = FusionDouble(AlternatingGroup(4))
                sage: sum(H.basis()).dual()
                Traceback (most recent call last):
                ...
                ValueError: dual is only available for simple objects
            """
            if not self.is_simple_object():
                raise ValueError("dual is only available for simple objects")
            return self.parent().dual(self)

        @cached_method
        def q_dimension(self, base_coercion=True):
            """
            Return the q-dimension of ``self``.

            This method is only available for simple objects.

            EXAMPLES::

                sage: G = AlternatingGroup(4)
                sage: H = FusionDouble(G)
                sage: [x.q_dimension() for x in H.basis()]
                [1, 1, 1, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4]
                sage: sum(x.q_dimension()^2 for x in H.basis()) == G.order()^2
                True

            TESTS::

                sage: H = FusionDouble(AlternatingGroup(4))
                sage: sum(H.basis()).q_dimension()
                Traceback (most recent call last):
                ...
                ValueError: quantum dimension is only available for simple objects
            """
            if not self.is_simple_object():
                raise ValueError("quantum dimension is only available for simple objects")
            return self.parent().s_ij(self, self.parent().one())
